(* Debparser *)
(* Copyright 2005-2007 Berke DURAK, INRIA Rocquencourt and The EDOS Project. *)
(* Released under the GNU LGPL version 2. *)

(** Parser for Debian package description files. *)

open Napkin;;

exception Error of string;;
exception Warning of string;;

let sf = Printf.sprintf;;
let fp = Printf.fprintf;;

(*** get_field *)
let rec get_field key = function
  | [] -> raise Not_found
  | (key',v1,v2)::rest ->
      if Util.lowercase_compare key key' then
        (v1,v2)
      else
        get_field key rest
;;
(* ***)
(*** get_field_first *)
let rec get_field_first key = function
  | [] -> raise Not_found
  | (key',v1,_)::rest ->
      if Util.lowercase_compare key key' then
        v1
      else
        get_field_first key rest
;;
(* ***)

(** Build a NAPKIN from the given row list. *)
let napkin_of_fields =
  let rex0 = Pcre.regexp ~study:true "\\s*(\\S+)\\s*"
  and rex1 = Pcre.regexp ~study:true "\\s*,\\s*"
  and rex2 = Pcre.regexp ~study:true "\\s*\\|\\s*"
  and rex3 = Pcre.regexp ~study:true "(\\S+)\\s*\\(\\s*(<|<<|<=|>=|>>|>|=)\\s*([0-9a-zA-Z.:+-~]+)\\s*\\)"
  and rex4 = Pcre.regexp ~study:true "^\\s*(\\S+)\\s*\\((\\S+)\\)$"
  in
  fun assoc ->
    let chump x =
      let ss = Pcre.exec ~rex:rex0 x in
      Pcre.get_substring ss 1
    in
    let (pk_unit, pk_version) =
      let u = chump (assoc "Package") in
      try
        let v = chump (assoc "Version") in
        (u,v)
      with
      | Not_found -> raise (Warning(sf "Package %S ignored because it has no version." u))
    in
    let get k =
      try
        Pcre.split ~rex:rex1 (assoc k)
      with
      | Not_found -> []
    in
    (*** Source *)
    let pk_source =
      try
        let ss = assoc "Source" in
        try
          let suv = Pcre.exec ~rex:rex4 ss in
          (Pcre.get_substring suv 1, Pcre.get_substring suv 2)
        with
        | Not_found -> (chump ss, pk_version)
      with
      | Not_found -> (pk_unit, pk_version)
    in
    (* ***)
    (*** Get an integer value *)
    let get_int64 field =
      try
        Int64.of_string (assoc field)
      with
      | Not_found|Invalid_argument _ -> 0L
    in
    (* ***)
    (*** Get a boolean value *)
    let get_bool field =
      try
        assoc field = "yes"
      with
      | Not_found -> false
    in
    (* ***)
    let pk_architecture = assoc "Architecture" in
    (*** Collect a conjunctive field *)
    let collect_conjunctive field =
      let stuff = get field in
      List.map (fun x ->
          try
            let ss = Pcre.exec ~rex:rex3 x in
            let (pn,op,v) = (Pcre.get_substring ss 1, Pcre.get_substring ss 2, Pcre.get_substring ss 3) in
            match op with
            | "<"     -> (pn, Sel_LT v) (* XXX *)
            | "<<"    -> (pn, Sel_LT v)
            | "<="    -> (pn, Sel_LEQ v)
            | "="     -> (pn, Sel_EQ v)
            | ">="    -> (pn, Sel_GEQ v)
            | ">"     -> (pn, Sel_GT v)
            | ">>"|_  -> (pn, Sel_GT v)
          with
          | Not_found -> (chump x,  Sel_ANY)) stuff
    in
    (* ***)
    (*** Collect disjunctive field *)
    let collect_disjunctive field =
      let stuff = get field in
      List.map (fun d ->
        let l = Pcre.split ~rex:rex2 d in
        List.map
          (fun x ->
            try
              let ss = Pcre.exec ~rex:rex3 x in
              let (pn,op,v) = (Pcre.get_substring ss 1, Pcre.get_substring ss 2, Pcre.get_substring ss 3) in
              match op with
              | "<"     -> (pn, Sel_LT v)
              | "<<"    -> (pn, Sel_LT v)
              | "<="    -> (pn, Sel_LEQ v)
              | "="     -> (pn, Sel_EQ v)
              | ">="    -> (pn, Sel_GEQ v)
              | ">"     -> (pn, Sel_GT v)
              | ">>"|_  -> (pn, Sel_GT v)
            with
            | Not_found ->
                (chump x, Sel_ANY)) l) stuff
    in
    (* ***)
    (*** Essential, build-essential *)
    { pk_extra           = ();
      pk_unit            = pk_unit;
      pk_version         = pk_version;
      pk_architecture    = pk_architecture;
      pk_source          = pk_source;
      pk_essential       = get_bool "Essential";
      pk_build_essential = get_bool "Build-Essential";
      pk_provides        = collect_conjunctive "Provides";
      pk_size            = get_int64 "Size";
      pk_installed_size  = get_int64 "Installed-Size";
      pk_conflicts       = collect_conjunctive "Conflicts";
      pk_replaces        = collect_conjunctive "Replaces";
      pk_depends         = collect_disjunctive "Depends";
      pk_pre_depends     = collect_disjunctive "Pre-Depends";
      pk_suggests        = collect_disjunctive "Suggests";
      pk_recommends      = collect_disjunctive "Recommends";
      pk_enhances        = collect_disjunctive "Enhances";
    }
;;
(* ***)
(*** napkin_of_row_list *)
let napkin_of_row_list rl = napkin_of_fields (fun key -> get_field_first key rl);;
(* ***)
(*** row_list_from_channel *)
let row_list_from_channel ?(tick=ignore) ?(warn=ignore) in_channel =
  let s = Stream.of_channel in_channel in
  let line = ref 0 in
  try
    let rs = Rfc822.row_stream_of_char_stream ~line s in
    let rec loop = parser
      | [< 'fl; s >] ->
          begin
            tick ();
            match
              try
                Some fl
              with
              | Warning w ->
                  warn w;
                  None
            with
            | None -> [< loop s >]
            | Some n -> [< 'n; loop s >]
          end
      | [< >] -> [< >]
    in
    loop rs
  with
  | Stream.Failure -> raise (Error(sf "Stream failure at line %d (position %d)" !line (Stream.count s)))
  | Stream.Error x -> raise (Error(sf "Stream error at line %d (position %d): %s" !line (Stream.count s) x))
;;
(* ***)
(*** from_channel *)
let from_channel ?tick ?warn in_channel =
  let rec loop = parser
    | [< 'fl; s >] -> [< '(napkin_of_row_list fl); loop s >]
    | [< >] -> [< >]
  in
  loop (row_list_from_channel ?tick ?warn in_channel)
;;
(* ***)
