(* Installability *)
(* Copyright 2005-2007 Berke DURAK, INRIA Rocquencourt and The EDOS Project. *)
(* Released under the GNU LGPL version 2. *)

open Napkin
open Rapids
open Diagnosis

module S = Solver.M(
  struct
    type t = (package_id, (unit_id, (version_number * release_number)) versioned) reason
    type reason = t
  end)
;;

module PS = Package_set;;
let sf = Printf.sprintf;;

(*** renumber_set *)
let renumber_set set =
  let m = PS.cardinal set in
  let backward = Hashtbl.create (2 * m) in
  let forward = Hashtbl.create (2 * m) in
  let i = ref 0 in
  PS.iter (fun p_id ->
    Hashtbl.add backward !i p_id;
    Hashtbl.add forward p_id !i;
    incr i) set;
  (backward, forward)
;;
(* ***)
let prepare log targets available db (indicator: Progress.indicator) =
  (* Compute the dependency closure of the set *)
  log "Computing closure";
  let set = targets in
  let set_closure = PS.union available (Functions.dependency_closure db targets) in
  let m = PS.cardinal set_closure in (* Number of affected packages *)
  log (sf "Done, %d packages in closure" m);
  indicator#start;
  log (sf "Numbering");
  (* Number packages *)
  let (backward, forward) = renumber_set set_closure in
  let f = Hashtbl.find forward
  and g = Hashtbl.find backward
  in
  (* Generate the boolean problem *)
  log (sf "Converting to boolean problem");
  let pb = S.initialize_problem m in
  let num_conflicts = ref 0
  and num_disjunctions = ref 0
  and num_dependencies = ref 0
  in
  let encoding_size = ref 0 in
  let charge x = encoding_size := !encoding_size + 2 in
  let j = ref 0 in
  (*** add conflicts and dependencies *)
  indicator#set_label "Conflicts and dependencies...";
  PS.iter
    begin fun p_id ->
      indicator#display (!j,m);
      incr j;
      let p = Functions.get_package_from_id db p_id in
			let pn = Unit_index.find (get_unit_index db) p.pk_unit in 
      (*** Add conflicts *)
			(* log (sf "Adding conflicts (%s)..." pn);*)
      (*let conflicts = PS.inter (get_conflicts db p_id) set_closure in*)
      let conflicts = PS.inter (Functions.conflicts db (PS.singleton p_id)) set_closure in
      PS.iter
        begin fun p_id' ->
          if p_id <> p_id' then
            begin
              let x = S.lit_of_var (f p_id) false
              and y = S.lit_of_var (f p_id') false
              in
              charge 2;
              S.add_bin_rule pb x y [Conflict(p_id, p_id')];
              incr num_conflicts
            end
        end
        conflicts;
      (* ***)
      (*** Add dependencies *)
			(* log (sf "Adding dependencies (%s)..." pn); *)
      let process conjunction =
        List.iter
          begin fun disjunction ->
            incr num_disjunctions;
            if disjunction = [] then
              () (* XXX : ignore empty disjunctions *)
            else
              begin
                let ps =
                  List.fold_left
                    begin fun ps versioned ->
                      incr num_dependencies;
											let (uid, sel) = versioned in
											(* log (sf "Selecting for unit %s" (Unit_index.find (get_unit_index db) uid)); *)
                      PS.union ps
                        (PS.inter available (Functions.select db versioned))
                    end
                    PS.empty
                    disjunction
                in
                (* let ps = PS.inter set ps in *)
                let o = PS.cardinal ps in
                if o = 0 then
                  begin
                    S.add_un_rule pb (S.lit_of_var (f p_id) false)
                      [Empty_disjunction(p_id, disjunction)]
                  end
                else
                  begin
                    let pl = PS.elements ps in
                    let vl = List.map f pl in
                    let l = List.map (fun v -> S.lit_of_var v true) vl in
                    charge (1 + o);
                    S.add_rule pb (Array.of_list ((S.lit_of_var (f p_id) false)::l))
                      [Dependency(p_id, pl)];
                    if l <> [] then S.associate_vars pb (S.lit_of_var (f p_id) true) vl
                  end
              end
          end
          conjunction
      in
      process p.pk_depends;
      process p.pk_pre_depends;
      (* ***)
    end
    set_closure;
  (* PS.iter
    begin fun p_id ->
      let v = S.lit_of_var (f p_id) false in
      S.add_un_rule pb v [Not_available p_id]
    end
    (PS.diff set_closure available); *)
  (* ***)
  (*let vars = List.map f (PS.elements set) in*)
  let failures = ref [] in
  log (sf "Done, formula of size %d" !encoding_size);
	(m, f, g, set, pb, failures, num_conflicts, num_disjunctions, num_dependencies);;

(*** check *)
let check
  db
  ?(log=ignore)
  ~(indicator : Progress.indicator)
  ~targets
  ~available
  ()
  =
	let (m, f, g, set, pb, failures, num_conflicts, num_disjunctions, num_dependencies) =
	prepare log targets available db indicator in
  indicator#set_label "Solving";
  let unknown = Array.make m true in
  for i = 0 to m - 1 do
    indicator#display (i,m);
    if PS.mem (g i) set && unknown.(i) then
      begin
        if S.solve pb i or (S.reset pb; S.solve pb i) then
          (* This package is OK *)
          begin
            let a = S.assignment pb in
            for j = 0 to m - 1 do
              if a.(j) = S.True then unknown.(j) <- false
            done
          end
        else
          begin
            S.reset pb;
            ignore (S.solve pb i);
            let reasons = S.collect_reasons pb i in
            failures := (g i, reasons) :: !failures;
            S.reset pb;
          end
      end
  done;
  indicator#finish;
  (* Sort failures *)
  let a = Array.of_list !failures in
  Array.sort
    begin fun (p_id,_) (p_id',_) ->
      let f p_id =
        let p = Functions.get_package_from_id db p_id in
        (Unit_index.find (get_unit_index db) p.pk_unit, p.pk_version)
      in
      compare (f p_id) (f p_id')
    end
    a;
  (* Create diagnosis *)
  { dg_closure_size = m;
    dg_conflicts = !num_conflicts;
    dg_disjunctions = !num_disjunctions;
    dg_dependencies = !num_dependencies;
    dg_failures = a }
;;
(*** check_archive *)
let check_archive 
	db
  ?(log=ignore)
  ~(indicator: Progress.indicator)
  ~(archive: PS.t)
  ~(available: PS.t) 
	()
	=
begin
	let nr_packages = Package_set.cardinal archive in
	let set_closure = PS.union available (Functions.dependency_closure db archive) in
	let m = PS.cardinal set_closure in
	let (backward, forward) = renumber_set set_closure in
	let fwd = Hashtbl.find forward
	and bwd = Hashtbl.find backward in
	let pb = S.initialize_problem m in
	PS.iter (fun p_id ->
		let p = Functions.get_package_from_id db p_id in
		let pn = Unit_index.find (get_unit_index db) p.pk_unit in
		(* conflicts *)
		PS.iter (fun p_id' ->
			if p_id <> p_id' then
				let x = S.lit_of_var (fwd p_id) false
				and y = S.lit_of_var (fwd p_id') false in
					S.add_bin_rule pb x y [Conflict(p_id, p_id')];
		) (PS.inter (Functions.conflicts db (PS.singleton p_id)) set_closure);
		(* dependencies *)
		let process conjunction =
		List.iter (fun disjunction ->
			if disjunction = [] then
				()
			else
			begin
				let ps = List.fold_left (fun ps versioned ->
					let (uid, sel) = versioned in
					PS.union ps (PS.inter available (Functions.select db versioned)))
				PS.empty
				disjunction in
				let o = PS.cardinal ps in
				if o = 0 then
					S.add_un_rule pb (S.lit_of_var (fwd p_id) false) [Empty_disjunction (p_id, disjunction)]
				else
				begin
					let pl = PS.elements ps in
					let vl = List.map fwd pl in
					let l = List.map (fun v -> S.lit_of_var v true) vl in
					S.add_rule pb (Array.of_list ((S.lit_of_var (fwd p_id) false)::l)) [Dependency(p_id, pl)];
					if l <> [] then S.associate_vars pb (S.lit_of_var (fwd p_id) true) vl
				end
			end) conjunction in
	process p.pk_depends;
	process p.pk_pre_depends) set_closure; 
	let failures = ref []
	and unknown = Array.make m true
	and installable = Array.make m false in
	for i = 0 to m - 1 do
		if PS.mem (bwd i) archive && unknown.(i) then
		begin
			if S.solve pb i or (S.reset pb; S.solve pb i) then
			(* This package is OK *)
			begin
				let a = S.assignment pb in
				for j = 0 to m - 1 do
					if a.(j) = S.True then (unknown.(j) <- false; installable.(j) <- true)
				done
			end
			else
			begin
				S.reset pb;
				ignore (S.solve pb i);
				let reasons = S.collect_reasons pb i in
				failures := (bwd i, reasons)::!failures;
				unknown.(i) <- false;
				installable.(i) <- false;
				S.reset pb;
			end
		end
	done;
	!failures
end;;
(* ***)
(*** install *)
(* let install
  db
  ?(log=ignore)
  ~(indicator : Progress.indicator)
  ~targets
  ~available
	?(known=PS.empty)
  ()
  =
	let (m, f, g, set, pb, failures, num_conflicts, num_disjunctions, num_dependencies) =
	prepare log targets available db indicator in
  indicator#set_label "Solving";
	let pkgs = ref PS.empty in
  let unknown = Array.make m true in
  for i = 0 to m - 1 do
    indicator#display (i,m);
		(* let p = Functions.get_package_from_id db (g i) in
		let pn = Unit_index.find (get_unit_index db) p.pk_unit in 
		prerr_endline ("Trying package " ^ pn); *)
    if PS.mem (g i) set && unknown.(i) then
		begin
       if S.solve pb i or (S.reset pb; S.solve pb i) then
         (* This package is OK *)
         begin
           let a = S.assignment pb in
           for j = 0 to m - 1 do
             if a.(j) = S.True then
						begin
							pkgs := PS.add (g j) !pkgs;
							unknown.(j) <- false
						end
           done
         end
		end
  done;
  indicator#finish;
	!pkgs
;; *)
(* ***)

(*** abundance *)
let abundance db ps =
begin
	let satisfied = ref true in
	PS.iter (fun p_id ->
		let p = Functions.get_package_from_id db p_id in
		List.iter (fun disj ->
			if !satisfied then
				let d_ok = List.exists
					(fun v -> let range = Functions.select db v in
						not (PS.is_empty (PS.inter ps range))
					) disj in
					if not d_ok then
						satisfied := false
		) (p.pk_depends @ p.pk_pre_depends);
	) ps;
	!satisfied	
end;;
(* ***)
